import matplotlib.pyplot as plt
import matplotlib.animation as animation
from matplotlib.lines import Line2D
import numpy as np
import math

def plot_obstacles(obstacles, ax):
    for obs in obstacles:
        ox, oy = obs.plot_obstacle()
        ax.scatter(ox, oy, s=7, c='k')

def animate_path_bug1(initial_robot_pos,goal_pos,path,obstacles):
    # Set plot variables
    fig = plt.figure()
    ax = fig.add_subplot(111, aspect='equal', autoscale_on=False,
                         xlim=(0, 50), ylim=(0, 60))
    ax.grid()

    goal, = ax.plot([], [], 'o', lw=2, c='g')
    path_line, = ax.plot([], [], 'r', lw=2)
    robot, = ax.plot([], [], 'ob', lw=3)
    plot_obstacles(obstacles, ax)
    steps_text = ax.text(0.02, 0.95, '', transform=ax.transAxes)

    # Define the start and goal positions 
    g_x, g_y = goal_pos
    s_x, s_y = initial_robot_pos
    
    if len(path) == 0:
        print("Path was not found!!")
        ani = 0
        return
        
    else:
        # Animation code
        def init():
            goal.set_data(g_x, g_y)
            robot.set_data(s_x, s_y)
            path_line.set_data(s_x, s_y)
            steps_text.set_text('')
            return robot, goal, path_line, steps_text

        def animate(i):
            """perform animation step"""
            if i < path.shape[0]:
                pos = path[i,:]
                robot.set_data(pos[0], pos[1])
                path_line.set_data(path[:i, 0], path[:i, 1])
                steps_text.set_text('Steps: %.1f' % i)
            return robot, path_line

        ani = animation.FuncAnimation(fig, animate, frames=600, repeat=False, interval=10, init_func=init)
        return ani


def animate_path_bug2(initial_robot_pos,goal_pos,path,obstacles):
    # Set plot variables
    fig = plt.figure()
    ax = fig.add_subplot(111, aspect='equal', autoscale_on=False,
                         xlim=(0, 50), ylim=(0, 60))
    ax.grid()

    goal, = ax.plot([], [], 'o', lw=2, c='g')
    line, = ax.plot([], [], 'r--', lw=1)
    path_line, = ax.plot([], [], 'g', lw=2)
    robot, = ax.plot([], [], 'ob', lw=4)
    steps_text = ax.text(0.02, 0.95, '', transform=ax.transAxes)
    custom_lines = [Line2D([0], [0], color='r', lw=4)]
    ax.legend(custom_lines, ['start-goal line'])

    plot_obstacles(obstacles, ax)

    s_x, s_y = initial_robot_pos
    g_x, g_y = goal_pos
    
    if len(path)==0:
        print("Path was not found!!")
    else:
        # Animation code
        def init():
            goal.set_data(g_x, g_y)
            robot.set_data(s_x, s_y)
            path_line.set_data(s_x, s_y)
            steps_text.set_text('')
            line.set_data([s_x, g_x], [s_y, g_y])
            return robot, goal, path_line, steps_text

        def animate(i):
            """perform animation step"""
            pos = path[i,:]
            robot.set_data(pos[0], pos[1])
            path_line.set_data(path[:i, 0], path[:i, 1])
            steps_text.set_text('Steps: %.1f' % i)
            return robot, path_line

        ani = animation.FuncAnimation(fig, animate, frames=700, blit=True, interval=10, init_func=init)
        return ani
    
def plot_circle(ax, x, y, size, color="-b"):  
        deg = list(range(0, 360, 5))
        deg.append(0)
        xl = [x + size * math.cos(np.deg2rad(d)) for d in deg]
        yl = [y + size * math.sin(np.deg2rad(d)) for d in deg]
        ax.plot(xl, yl, color)    
    
def animate_path_rrt(rrt):
    path = rrt.planning()
    
    fig = plt.figure(figsize=(8, 6))
    ax = fig.add_subplot(111, aspect='equal', autoscale_on=False,
                         xlim=(0, 16), ylim=(0, 12))
    ax.grid()

    ax.plot(rrt.start.x, rrt.start.y, "^r", lw=5)
    ax.plot(rrt.end.x, rrt.end.y, "^c", lw=5)

    for obs in rrt.obstacle_list:
        cx, cy = obs.center
        plot_circle(ax, cx, cy, obs.radius)

    for node in rrt.node_list:
        if node.parent:
            ax.plot(node.path_x, node.path_y, "g.-")

    line_path, = ax.plot([], [], "b", lw=3)

    if path is not None:
        path_in_order = np.flipud(path)
        start_pos = path_in_order[0,:]

        def init():
            line_path.set_data(path_in_order[0], path_in_order[1])
            return line_path

        def animate(i):
            """perform animation step"""
            line_path.set_data(path_in_order[:i,0], path_in_order[:i, 1])
            return line_path

        ani = animation.FuncAnimation(fig, animate, frames=30, blit=True, interval=100, init_func=init)
        return ani
    else:
        print("Path was not found!!")
    
def animate_path_rrtc(rrtc):
    path = rrtc.planning()
    
    fig = plt.figure(figsize=(8, 6))
    ax = fig.add_subplot(111, aspect='equal', autoscale_on=False,
                         xlim=(0, 16), ylim=(0, 12))
    ax.grid()

    ax.plot(rrtc.start.x, rrtc.start.y, "^r", lw=5)
    ax.plot(rrtc.end.x, rrtc.end.y, "^c", lw=5)

    for obs in rrtc.obstacle_list:
        cx, cy = obs.center
        plot_circle(ax, cx, cy, obs.radius)

    for node in rrtc.start_node_list:
        if node.parent:
            ax.plot(node.path_x, node.path_y, "g.-")

    for node in rrtc.end_node_list:
        if node.parent:
            ax.plot(node.path_x, node.path_y, "b.-")
            
    line_path, = ax.plot([], [], "r", lw=3)
    custom_lines = [Line2D([0], [0], color='g', lw=4), Line2D([0], [0], color='b', lw=4)]
    ax.legend(custom_lines, ['Start tree', 'End tree'])

    if path is not None:
        path = np.array(path)
        start_pos = path[0,:]
        
        def init():
            line_path.set_data(start_pos[0], start_pos[1])
            return line_path

        def animate(i):
            """perform animation step"""
            line_path.set_data(path[:i,0], path[:i, 1])
            return line_path

        ani = animation.FuncAnimation(fig, animate, frames=30, repeat=False, interval=100, init_func=init)
        return ani
    else:
        print("Path was not found!!")
        
def animate_path_prm(rmap,start, goal, num_samples, robot_size, max_distance, max_neighbours):
    fig = plt.figure(figsize=(8, 6))
    ax = fig.add_subplot(111, aspect='equal', autoscale_on=False,
                         xlim=(-10, 70), ylim=(-10, 70))
    plt.grid(True)

    obs = rmap.obstacles.data
    ax.plot(obs[:,0], obs[:,1], ".k")

    rmap.__generate_roadmap__(num_samples, max_distance, max_neighbours, robot_size)

    for i, v_edges in enumerate(rmap.edges):
        for e_idx in v_edges:
            v_from = rmap.vertices[i,:]
            v_to = rmap.vertices[e_idx,:]

            ax.plot([v_from[0], v_to[0]],
                    [v_from[1], v_to[1]], "k--", lw=0.5)
    
    path = rmap.plan(start, goal)

    # Animation code
    start_plot, = ax.plot([], [], "^r", lw=5)
    goal_plot, = ax.plot([], [], "^c", lw=5)
    line_path, = ax.plot([], [], "b", lw=2)    

    if path is not None:
        path_in_order = np.flipud(path)
        start_pos = path_in_order[0,:]

        def init():
            print(init)
            start_plot.set_data(start[0], start[1])
            goal_plot.set_data(goal[0], goal[1])
            line_path.set_data([], [])
            return line_path, start_plot, goal_plot

        def animate(i):
            """perform animation step"""
            line_path.set_data(path_in_order[:i,0], path_in_order[:i, 1])
            return line_path

        ani = animation.FuncAnimation(fig, animate, frames=30, blit=True, interval=30, init_func=init)
        return ani
    else:
        print("Path was not found!!")


import numpy as np
from .math_functions import *


class Polygon:
    """
    Obstacles are represented as polygons
    Polygons are defined as an array with n rows (vertices) and 2 columns
    
    """

    def __init__(self, vertices=np.zeros((4,2))):
        self.vertices = vertices

    def compute_distance_point_to_polygon(self, point_q, ccw):
        """
        Compute distance from point_q to the closest point in the polygon

        Method returns:
        - dist: minimal distance from point_q to polygon
        - indices of segment closest to point_q

        """        
        dist = np.inf
        segment_idx = None
        closest_point = None
        len_polygon = self.vertices.shape[0]
        
        if ccw:
            for i in range(len_polygon):
                case, seg_dist, _ = compute_distance_point_to_segment(self.vertices[(i+1) % len_polygon],self.vertices[i], point_q)
                if seg_dist <= dist:
                    dist = seg_dist
                    if case == 0:
                        closest_point = (i+1) % len_polygon
                    elif case == 1:
                        closest_point = (i+1) % len_polygon
                    else:
                        closest_point = i  

            segment_idx = (closest_point, (closest_point+len_polygon-1) % len_polygon)
            # print("Closest segment is {}, {}".format(segment_idx[0], segment_idx[1]))
            
        else:
            for i in range(len_polygon):
                case, seg_dist, _ = compute_distance_point_to_segment(self.vertices[i],self.vertices[(i+1) % len_polygon], point_q)
                if seg_dist <= dist:
                    dist = seg_dist
                    if case == 0:
                        closest_point = i
                    elif case == 1:
                        closest_point = i
                    else:
                        closest_point = (i+1) % len_polygon   

            segment_idx = (closest_point, (closest_point+1) % len_polygon)
            # print("Closest segment is {}, {}".format(segment_idx[0], segment_idx[1]))
        return dist, segment_idx

    def compute_tangent_vector_to_polygon(self, point_q, idx):  
        
        """
        Determines the unit-length vector tangent at point_q to the polygon
        
        Method returns:
           tangent vector

        """  
 
        v1 = self.vertices[idx[0]]
        v2 = self.vertices[idx[1]]
        
        case, seg_dist, closest_point = compute_distance_point_to_segment(v1, v2, point_q)
        
        tangent_vector = (v2-v1)/np.linalg.norm(v2-v1)
        
        return tangent_vector


    def plot_obstacle(self):
        ox = []
        oy = []

        for i in range(self.vertices.shape[0]):
            x0, y0 = self.vertices[i]
            x1, y1 = self.vertices[(i+1) % self.vertices.shape[0]]

            if x0 == x1:
                start = np.min([y0, y1])
                end = np.max([y0, y1])
                y_values = np.linspace(start, end, int(np.ceil((end-start)/0.2)), endpoint=True) 
                x_values = [x0] * len(y_values)
                ox.extend(x_values)
                oy.extend(y_values)
                
            if y0 == y1:
                start = np.min([x0, x1])
                end = np.max([x0, x1])
                x_values = np.linspace(start, end, int(np.ceil((end-start)/0.2)), endpoint=True) 
                y_values = [y0] * len(x_values)
                ox.extend(x_values)
                oy.extend(y_values)

        return ox, oy


    def to_display_format(self, screen_height):
        coordinates = [coordinates_to_pygame(v, screen_height) for v in self.vertices[0:-1]]
        return coordinates


    def is_in_collision_with_points(self, points, min_dist=2.5):
        # First check if point is within polygon
        points_in_collision = []
        
        for point in points:
            count_collisions = 0
            p_x, p_y = point
            for i in range(self.vertices.shape[0]-2):
                j = i - 1 if i != 0 else self.vertices.shape[0]-2
                v1 = self.vertices[i]
                v2 = self.vertices[j]

                if ((v1[1] > p_y) != (v2[1] > p_y)) and (p_x < (v2[0] - v1[0]) * (p_y - v1[1]) / (v2[1] - v1[1] + v1[0])):
                    count_collisions += 1

            if count_collisions % 2 == 1:
                Qpoints_in_collision.append(point)

        if len(points_in_collision):
            return True

        # Second check if point is in collision with edges
        dist, _ = self.compute_distance_point_to_polygon(points[-1])
        if dist < min_dist:
            return True

        return False


    def get_perimeter(self):

        perimeter = 0

        for i in range(self.vertices.shape[0]-1):
            v1 = self.vertices[i]
            v2 = self.vertices[i+1]

            perimeter += compute_distance_between_points(v1, v2)

        return perimeter


class Rectangle(Polygon):

    def __init__(self, origin=np.zeros(2), width=100, height=20):
        self.width = width
        self.height = height
        self.origin = origin

        v1 = origin
        v2 = origin + np.array([width, 0])
        v3 = origin + np.array([width, -height])
        v4 = origin + np.array([0, -height])

        Polygon.__init__(self, vertices=np.array([v1, v2, v3, v4]))

    def to_display_format(self, screen_height):
        py_origin = coordinates_to_pygame(self.origin, screen_height)
        return (py_origin[0], py_origin[1], self.width, self.height)

    def plot_obstacle(self):
        return super().plot_obstacle()


class Circle:
    
    def __init__(self, c_x, c_y, radius):
        self.center = np.array([c_x, c_y])
        self.radius = radius

    def is_in_collision_with_points(self, points):

        dist = []
        for point in points:
            dx = self.center[0] - point[0]
            dy = self.center[1] - point[1]

            dist.append(dx * dx + dy * dy)

        if np.min(dist) <= self.radius ** 2:
            return True

        return False  # safe


import numpy as np
import math

def compute_distance_between_points(p1, p2):
    """ 
        Computes distance between two points
    """
    dx = p2[0] - p1[0]
    dy = p2[1] - p1[1]

    return math.hypot(dx, dy)


def is_point_in_segment(start_seg, end_seg, point_q):
    """ 
        Determines in point_q is strictly inside the segment defined by start_seg and end_seg
    """
    dist_1 = compute_distance_between_points(start_seg, point_q)
    dist_2 = compute_distance_between_points(end_seg, point_q)
    dist_3 = compute_distance_between_points(start_seg, end_seg)
     
    return dist_1 + dist_2 == dist_3


def compute_lines_intersection(line_1, line_2):
    
    """ 
    The orthogonal projection of a point onto a line is defined as the
    intersection between two perpendicular lines (with the point of
    interest being along one of the lines)
    
    This orthogonal projection can be obtained by finding the inteserction
    point between 2 perpendicular lines
       
    Lines are defined in standard form ax + by + c = 0
    line_1 = <a, b, c>
    line_2 = <a1, b1, c1>
    """
    d  = line_1[0] * line_2[1] - line_1[1] * line_2[0]
    dx = line_1[2] * line_2[1] - line_1[1] * line_2[2]
    dy = line_1[0] * line_2[2] - line_1[2] * line_2[0]
    if d != 0:
        x = dx / d
        y = dy / d
        return np.array([x,y])
    return False


def compute_line_through_points(p1, p2):
   
    """ 
    Computes line defined by 2 points
    Line is returned in standard form ax + by + c = 0
    """
    p1 = np.array(p1)
    p2 = np.array(p2)
    
    # If p1 and p2 are different, compute the line defined by these 2 points
    if np.all(np.isclose(p1, p2)):
        return False
    
    a = p1[1] - p2[1]
    b = p2[0] - p1[0]
    c = p1[0]*p2[1] - p2[0]*p1[1]

    return [a, b, -c]


def compute_distance_point_to_line_by_intersection(start_line, end_line, point_q):
    
    """
    Computes distance from point_q and the line defined by 
    (start_line, end_line).
    
    This corresponds to finding the closest point (x0,y0) on the line to point_q
    (i.e, orthogonal projection of point_q onto the line) and computing
    the distance between (x0,y0) and point_q
    
    (x0, y0) coordinates are computed by considering the intersection between
    two perpendicular lines. The first line is defined by the points start_line
    and end_line. The second line is perpendicular to the first one and goes through
    point_q
    
    """

    # Put point_q in right format
    point_q = np.array(point_q)
    
    # Compute first line parameters
    first_line = compute_line_through_points(start_line, end_line)
           
    if first_line:
        a, b, c = first_line
        
        # Normalize parameters so that a*a + b*b = 1
        ab_norm = np.linalg.norm(np.array([a, b]))
        a_norm = a / ab_norm
        b_norm = b / ab_norm
        
        # Define a norm vector perpendicular to first line
        # This vector is parallel to the line that goes through point_q
        normal = np.array([a_norm, b_norm])

        # Using the vector-form equation of a line, to find a second in line 2
        point_q2 = point_q + 2*(normal)
        
        # Using point_q and point_q2, find the standard parameters of second line
        second_line = compute_line_through_points(point_q, point_q2)
        if second_line:

            # Compute intersection point
            proj_point = compute_lines_intersection(first_line, second_line)
            
            if proj_point is not None:
                # Compute distance between point_q and its projection onto first line
                distance = compute_distance_between_points(proj_point, point_q)
                return proj_point, distance
            
    return False


def compute_distance_point_to_segment(start_seg, end_seg, point_q):
    """
    Computes distance from point_q and line segment defined by start_seg and end_seg
    
    This method first computes the distance (and othorgonal projection of point_q) 
    to the line defined by start_seg and end_seg.
    
    If proj_point_q is stricly inside the segment, it returns the distance
    to the line and the indicator w=1
    
    If proj_point_q is not in the segment, it determines the closest segment
    point to point_q, computes the distance between point_q and the chosen segment point
    
    The indicator will be set to w=1 if start_seg is the closest. Otherwise,
    end_seg is the closest point and w=2
    
    """
    
    # Compute point_q projection and distance from point_q to projection
    proj_q, distance = compute_distance_point_to_line_by_intersection(start_seg, end_seg, point_q)
    
    # Compute distance to start and end of segment
    dist_to_start = compute_distance_between_points(start_seg, point_q)
    dist_to_end = compute_distance_between_points(end_seg, point_q)
    
    if is_point_in_segment(start_seg, end_seg, proj_q):
        w = 0  # orthogonal projection is the closest point
    elif dist_to_start < dist_to_end:
        w = 1 # start is the closest point in segment
        distance = dist_to_start
        proj_q = start_seg
    else:
        w = 2 # end is the closest point in segment
        distance = dist_to_end
        proj_q = end_seg
                
    return w, distance, proj_q


def get_direction_from_points(p1, p2):
    """
    Computes horizontal angle between line defined by p1 and p2 and world x-axis
    """
    dx = p2[0] - p1[0]
    dy = p2[1] - p1[1]
    angle = math.atan2(dy, dx)
    return np.array([np.cos(angle), np.sin(angle)])


def get_direction_from_line(line):
    """
    Computes direction of vector parallel to standard line ax + by + c = 0
    """
    a, b, c = line
    vector = np.array([b, -a])
    vector = vector / np.linalg.norm(vector)
    return vector


def point_in_line(line, point):
    """
    Determines if a point is strictly inside a line
    """
    a, b, c = line
    return np.isclose(a*point[0] + b*point[1] - c, 0)


def find_nearest(array, value):
    array = np.asarray(array)
    idx = (np.linalg.norm(array - value, axis=1)).argmin()
    return idx, array[idx]


import matplotlib.pyplot as plt
import ipywidgets as widgets
import io
import numpy as np
import meshcat.geometry as g

def intro(vis):
    secret_msg = ["Hello", "students", "welcome", "to", "ECE4078"]
    w_slider = widgets.IntSlider(value=0, min=0, max=len(secret_msg)-1, step=1, description='Omega',
                                                continuous_update=False)
    
    size = 8
    def f(change):
        vis["render"].set_object(g.SceneText(secret_msg[change], width=size, height=size, font_size = size))
    interactive_plot = widgets.interactive_output(f, {'change': w_slider})
    return interactive_plot, w_slider


def NotebookChecker(vis):
    w_slider = widgets.IntSlider(value=0, min=0, max=10, step=1, description='Omega',
                                                continuous_update=False)
    def f(change):
        fig = plt.figure(figsize=(5, 5))
        ax = plt.gca()

        x = np.linspace(1, 2 * np.pi, 100)
        y = np.sin(x*change)
        ax.plot(x, y)
        image_data = io.BytesIO()
        fig.savefig(image_data, bbox_inches = 'tight', 
                        facecolor='white', transparent=False)
        plt.close()
        size = 8
        if change == 10:
            vis.delete()
            vis["render"].set_object(g.SceneText("Isaac Asimov",
                                            width=size,
                                            height=size,
                                            font_size = size))
        else:
            vis["render"].set_object(
            g.Plane(width=1, height=1),
            g.MeshPhongMaterial(
                map=g.ImageTexture(image=g.PngImage(image_data.getvalue())), 
                transparent=True, needsUpdate=True))

    interactive_plot = widgets.interactive_output(f, {'change': w_slider})
    return interactive_plot, w_slider


from setuptools import setup, find_packages

VERSION = '1.1.2'
DESCRIPTION = "Support file for ece4078 practicals content"


setup(
    name="ece4078",
    version=VERSION,
    author="Tin Tran",
    author_email="tin.tran1@monash.edu",
    description=DESCRIPTION,
    packages=find_packages("src"),
    package_dir={"": "src"},
    package_data={'': ['install_nginx', 'nginx-meshcat-proxy.conf', '*.pyc']},
    install_requires=[],
    include_package_data=True
)



import os
import numpy as np
import matplotlib.pyplot as plt
from matplotlib.patches import FancyArrowPatch
from matplotlib.patches import Ellipse
import matplotlib.transforms as transforms
from ipywidgets import interact, widgets, Layout, Button, Box, VBox, IntSlider
import threading as thrd
import time
from matplotlib.animation import FuncAnimation

class Render_class():
    def init_plot(self):

        # Initialize figure
        fig = plt.figure(figsize=(6, 6))
        ax = plt.gca()
        ax.set_xlim([-4,-1.5])
        ax.set_ylim([-3.5,-1.5])
        ax.tick_params(axis='both', which='major', labelsize=7)
        plt.title('Overhead View')
        plt.xlabel('X (m)',weight='bold')
        plt.ylabel('Y (m)',weight='bold')

        self.figure = fig

         # Plot ground truth trajectory
        ax.plot(self.true_state[:, 0], self.true_state[:, 1], c='r', lw=1, label='True state (from data)')
        
        # Plot current model state
        self.line, = ax.plot(self.state[0,0], self.state[0, 1], c='b', lw=1, label='Predicted state (model)')

            
        ax.legend(loc='upper right')
        # Create Robot Axes 
        self.robot_ax = []
        self.robot_ax.append(FancyArrowPatch((0,0), (0.15,0),
                                            mutation_scale=8,color='red'))
        self.robot_ax.append(FancyArrowPatch((0,0), (0,0.15),
                                            mutation_scale=8,color='green'))
        
        # Apply translation and rotation as specified by current robot state
        cos_theta, sin_theta = np.cos(self.state[0,2]), np.sin(self.state[0,2])
        Tw_r = np.eye(3)
        Tw_r[0:2,2] = self.state[0,:2]
        Tw_r[0:2,0:2] = [[cos_theta,-sin_theta],[sin_theta,cos_theta]]
        Tw_r_obj = transforms.Affine2D(Tw_r)
        self.ax_trans = ax.transData
        self.robot_ax[0].set_transform(Tw_r_obj+self.ax_trans)
        self.robot_ax[1].set_transform(self.robot_ax[0].get_transform())
        ax.add_patch(self.robot_ax[0])
        ax.add_patch(self.robot_ax[1])


        if self.add_aruco:
            marker_files = [filename for filename in os.listdir('Practical03_Support/images') if filename.startswith("M")]
            marker_world_width = 0.3
            for i,filename in enumerate(sorted(marker_files)):
                fprts = filename.split('_')
                mp = np.array([float(fprts[1]),float(fprts[2])])
                mi = plt.imread('Practical03_Support/images/'+filename)
                ext = [mp[0]-marker_world_width/2,mp[0]+marker_world_width/2,\
                mp[1]-marker_world_width/2,mp[1]+marker_world_width/2]
                plt.imshow(mi,extent=ext)
                ax.annotate(str(i),(mp[0],mp[1]),color='red',weight='bold')


        #Set up to plot measurements
        if self.measurements is not None:
            self.marker_lines = []
            self.marker_lables = []
            self.marker_scatter = plt.scatter(np.zeros((10,1)),np.zeros((10,1)),color='lime')
            for i in range(10):
                ln, = ax.plot(np.zeros((2,1)),np.zeros((2,1)),color='yellow',alpha=0.5)
                self.marker_lines.append(ln)
                an = ax.annotate(str(i),(-3,-2.5),color='green',weight='bold')
                self.marker_lables.append(an)
                
        #Set up to plot added landmarks
        if self.landmarks is not None:
            self.landmarks_scatter = ax.scatter(self.landmarks[0][0,:],self.landmarks[0][1,:],s=80,color='red')
        
        if self.marker_cov is not None:
            num_landmarks = self.landmarks[-1].shape[1]
            self.marker_ells = []
            for i in range(num_landmarks):
                el = Ellipse((0,0),
                              width=0.3, height=0.3,
                              angle=0,facecolor='none',edgecolor='blue')
                self.marker_ells.append(el)
                ax.add_patch(el)
            for i in range(self.landmarks[self.cur_frame].shape[1]):
                self.marker_ells[i].set_center((self.landmarks[self.cur_frame][0,i],self.landmarks[self.cur_frame][1,i]))
        
        #Set up ellipsoid to draw robot covariance
        if self.robot_cov is not None:
            cov = self.robot_cov[0,:,:]
            a = cov[0,0]
            b = cov[0,1]
            c = cov[1,0]
            d = cov[1,1]
            B = -(a+d)
            C = (a*d-b*c)
            lam1 = (-B+np.sqrt(B**2-4*C))/2
            lam2 = (-B-np.sqrt(B**2-4*C))/2
            v1 = np.array([[lam1-d],[c]])
            self.robo_ell = Ellipse((self.state[0,0],self.state[0,1]),
                              width=lam1, height=lam2,
                              angle=np.rad2deg(np.arccos(v1[0]/np.linalg.norm(v1))))
            self.robo_ell.set_facecolor('none')
            self.robo_ell.set_edgecolor('blue')
            ax.add_patch(self.robo_ell)


    def step_render(self):
        self.line.set_data(self.state[0:self.cur_frame,0],self.state[0:self.cur_frame,1])
        self.figure.canvas.draw_idle()
        
        c, s = np.cos(self.state[self.cur_frame,2]), np.sin(self.state[self.cur_frame,2])
        Tw_r = np.eye(3)
        Tw_r[0:2,2] = [self.state[self.cur_frame,0],self.state[self.cur_frame,1]]
        Tw_r[0:2,0:2] = [[c,-s],[s,c]]
        Tw_r_obj = transforms.Affine2D(Tw_r)
        self.robot_ax[0].set_transform(Tw_r_obj+self.ax_trans)
        self.robot_ax[1].set_transform(self.robot_ax[0].get_transform())
        
        #Render Measurements
        if self.measurements is not None:
            # Construct a 2x2 rotation matrix from the robot to world
            th = self.state[self.cur_frame,2]
            Rot_0_rob = np.block([[np.cos(th), -np.sin(th)],[np.sin(th), np.cos(th)]])
            robot_xy = self.state[self.cur_frame,0:2].reshape(-1,1)
            for i in range(len(self.marker_lines)):
                self.marker_lines[i].set_visible(False)
                self.marker_lables[i].set_visible(False)
            marker_pos_all = np.zeros((len(self.measurements[self.cur_frame]),2))
            for i in range(len(self.measurements[self.cur_frame])):
                mes = self.measurements[self.cur_frame][i]
                marker_pos = Rot_0_rob.dot(mes.position.reshape(-1,1)) + robot_xy
                marker_pos_all[i,0] = marker_pos[0]
                marker_pos_all[i,1] = marker_pos[1]
                self.marker_lines[mes.tag].set_data([robot_xy[0],marker_pos[0]],[robot_xy[1],marker_pos[1]])
                self.marker_lines[mes.tag].set_visible(True)
                self.marker_lables[mes.tag].set_x(marker_pos[0])
                self.marker_lables[mes.tag].set_y(marker_pos[1])
                self.marker_lables[mes.tag].set_visible(True)
            self.marker_scatter.set_offsets(marker_pos_all)
        
        #Render Landmarks
        if self.landmarks is not None:
            self.landmarks_scatter.set_offsets(self.landmarks[self.cur_frame].transpose())
        
        #Render Robot Covariance Ellipse
        if self.robot_cov is not None:
            cov = self.robot_cov[self.cur_frame,:,:]
            a = cov[0,0]
            b = cov[0,1]
            c = cov[1,0]
            d = cov[1,1]
            B = -(a+d)
            C = (a*d-b*c)
            lam1 = (-B+np.sqrt(B**2-4*C))/2
            lam2 = (-B-np.sqrt(B**2-4*C))/2
            v1 = np.array([[lam1-d],[c]])
            self.robo_ell.width = lam1
            self.robo_ell.height = lam2
            self.robo_ell.angle = np.rad2deg(np.arccos(v1[0]/np.linalg.norm(v1)))
            self.robo_ell.set_center((self.state[self.cur_frame,0],self.state[self.cur_frame,1]))
        if self.marker_cov is not None:
            for i in range(self.landmarks[self.cur_frame].shape[1]):
                self.marker_ells[i].set_center((self.landmarks[self.cur_frame][0,i],self.landmarks[self.cur_frame][1,i]))
                cov = self.marker_cov[self.cur_frame][i*2:i*2+2,i*2:i*2+2]
                a = cov[0,0]
                b = cov[0,1]
                c = cov[1,0]
                d = cov[1,1]
                B = -(a+d)
                C = (a*d-b*c)
                lam1 = (-B+np.sqrt(B**2-4*C))/2
                lam2 = (-B-np.sqrt(B**2-4*C))/2
                v1 = np.array([[lam1-d],[c]])
                self.marker_ells[i].width=lam1
                self.marker_ells[i].height=lam2
                self.marker_ells[i].angle = np.rad2deg(np.arccos(v1[0]/np.linalg.norm(v1)))

class video_Render(Render_class):
    def __init__(self, model_state, measurements=None, true_state=None, landmarks=None, robot_cov=None,
        marker_cov=None, add_aruco=False):

        self.measurements = measurements
        self.landmarks = landmarks
        self.marker_cov = marker_cov
        self.true_state = true_state
        self.robot_cov = robot_cov        
        self.state = model_state
        self.add_aruco = add_aruco
        self.init_plot()
        plt.close()
        
    def render_video(self):
        def render(frame):
            self.cur_frame = frame
            self.step_render()
        anim = FuncAnimation(self.figure, render, frames=self.state.shape[0], interval=200)
        return anim


class Singleton:
    def __init__(self, cls):
        self._cls = cls

    def Instance(self):
        try:
            return self._instance
        except AttributeError:
            self._instance = self._cls()
            return self._instance

    def __call__(self):
        raise TypeError('Singletons must be accessed through `Instance()`.')

    def __instancecheck__(self, inst):
        return isinstance(inst, self._cls)

@Singleton
class Renderer(thrd.Thread, Render_class):
    
    #Make singleton
    _instance = None
    
    def __init__(self):
        # Call the Thread class's init function
        thrd.Thread.__init__(self)

        
    def initialize(self, model_state, measurements=None, true_state=None, landmarks=None, robot_cov=None,
        marker_cov=None, add_aruco=False, dt_data=0.2):
        self.lock = thrd.Lock()

        self.initialized = False
        self.measurements = measurements
        self.landmarks = landmarks
        self.marker_cov = marker_cov
        self.true_state = true_state
        self.robot_cov = robot_cov        
        self.paused = False
        self.cur_frame = 0
        self.dt_data = dt_data
        self.dt_render = dt_data
        self.state = model_state
        self.add_aruco = add_aruco

        self.init_plot()

        btn_play = widgets.Button(description='Play/Pause', layout=Layout(flex='1 1 0%', width='auto'), button_style='success')
        btn_play.on_click(self.pause)
        
        btn_prev = widgets.Button(description='<<', layout=Layout(flex='0.3 1 0%', width='auto'), button_style='warning')
        btn_prev.on_click(self.prv)
        
        btn_next = widgets.Button(description='>>', layout=Layout(flex='0.3 1 0%', width='auto'), button_style='warning')
        btn_next.on_click(self.nxt)

        controls = [
            IntSlider(description='Frame: ', layout=Layout(flex='3 1 0%', width='auto'),min=0, max=(model_state.shape[0]-1)),
            btn_prev,
            btn_play,
            btn_next
         ]
    
    
        self.slider = controls[0]
        self.slider.observe(self.slider_change, names='value')
        
        box_layout = Layout(display='flex',
                            flex_flow='row',
                            align_items='stretch',
                            width='70%')
        display(Box(children=controls, layout=box_layout))
            
        if not self.is_alive():
                self.start()
                
        self.initialized = True
                            
    #Render Loop
    def run(self):
        while True:
            if self.paused == False:
                self.cur_frame = int(self.cur_frame + self.dt_render/self.dt_data)
                if self.cur_frame >= self.state.shape[0]:
                    self.cur_frame = 0
                if self.initialized == True:
                    self.render()
            time.sleep(self.dt_render)       
            
    def render(self):
        self.lock.acquire()
        self.slider.value = self.cur_frame
        self.step_render()
        self.lock.release()
        
    def pause(self,b=None):
        self.paused = not self.paused
    
    def prv(self,b=None):
        self.paused = True
        self.cur_frame = int(self.cur_frame-1)
        self.slider.value = self.cur_frame
        if self.cur_frame < 0:
            self.cur_frame = 0
        self.render()
    
    def nxt(self,b=None):
        self.paused = True
        self.cur_frame = int(self.cur_frame + 1)
        self.slider.value = self.cur_frame
        if self.cur_frame >= self.state.shape[0]:
            self.cur_frame = self.state.shape[0]-1
        self.render()
    
    def slider_change(self,change):
        if self.paused == True:
            self.cur_frame = change['new']
            self.render()
